/*
 * Copyright (c) 2009 Michal Sarnovsky
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <inttypes.h>
#include <string.h>
#include <stdbool.h>

#include "uart0_config.h"

#ifndef __UART__
#define __UART__


/**
 * \addtogroup uart0
 *
 * @{
 */
/**
 * \file
 * UART0 header
 *
 * \author Michal Sarnovsky
 */
/**
 * \addtogroup uart0_read
 */
/**
 * @}
 */


#if defined(__AVR_ATmega644__)
#elif defined(__AVR_ATmega48__)
  #define USART0_RX_vect USART_RX_vect
  #define USART0_UDRE_vect USART_UDRE_vect
#elif defined(__AVR_ATmega8__)
  #define USART0_RX_vect USART_RXC_vect
  #define USART0_UDRE_vect USART_UDRE_vect
  #define UDR0 UDR
  #define UCSR0A UCSRA
    #define RXC0 RXC
    #define TXC0 TXC
    #define UDRE0 UDRE
    #define FE0 FE
    #define DOR0 DOR
    #define UPE0 PE
    #define U2X0 U2X
    #define MPCM0 MPCM
  #define UCSR0B UCSRB
    #define RXCIE0 RXCIE
    #define TXCIE0 TXCIE
    #define UDRIE0 UDRIE
    #define RXEN0 RXEN
    #define TXEN0 TXEN
    #define UCSZ02 UCSZ2
    #define RXB80 RXB8
    #define TXB80 TXB
  #define UCSR0C UCSRC
    #define UPM00 UPM0
    #define UPM01 UPM1
    #define USBS0 USBS
    #define UCSZ01 UCSZ1
    #define UCSZ00 UCSZ0
    #define UCPOL0 UCPOL
  #define UBRR0H UBRRH
  #define UBRR0L UBRRL
#else
  #error Select supported hardware platform. This driver supports: ATMEGA644, ATMEGA8.
#endif

#ifndef F_CPU
  #error No clock for UART0 defined (define F_CPU value [Hz]).
#endif

#ifndef UART0_BAUDRATE
  #error No baudrate for UART0 defined (define UART0_BAUDRATE in config.h file).
#endif

#ifndef UART0_RX_CALLBACK_SUPPORT
  #error No RX callback support defined (define UART0_RX_CALLBACK_SUPPORT in config.h file).
#endif

#ifndef UART0_RX_BUFFER_SIZE
  #error No RX buffer size defined (define UART0_RX_BUFFER_SIZE in config.h file).
#else

  #if UART0_RX_BUFFER_SIZE < 2
    #error UART0_RX_BUFFER_SIZE must at least 2.
  #endif
#endif

#ifndef UART0_TX_BUFFER_SIZE
  #error No TX buffer size defined (define UART0_TX_BUFFER_SIZE in config.h file).
#else
  #if UART0_TX_BUFFER_SIZE < 2
    #error UART0_TX_BUFFER_SIZE must at least 2.
  #endif
#endif

#ifndef UART0_RX_OVERFLOW_REWRITE
  #error No RX buffer overflow rewrite support defined (define UART0_RX_OVERFLOW_REWRITE in config.h file).
#endif


/*--------------------------------------------------------------------*/


/**
 * \ingroup uart0
 * Frame format: asynchronous, no parity, one stop bit, 8-bit character size
 */
#if defined(__AVR_ATmega644__) || defined(__AVR_ATmega48__)
#define UCSR0C_VALUE ((0 << UMSEL01) | (0 << UMSEL00) | (0 << UPM01) | (0 << UPM00) | \
                      (0 << USBS0) | (1 << UCSZ01) | (1 << UCSZ00))
#endif
#if defined(__AVR_ATmega8__)
#define UCSR0C_VALUE ((0 << UMSEL) | (0 << UPM01) | (0 << UPM00) | (0 << USBS0) | (1 << UCSZ01) | \
                      (1 << UCSZ00))
#endif

#define UBRR0_MULTIPLIER 850 /* multiply constant */
#define UART0_ALLOWED_DEVIATION_MUL 34 /* 4 % is allowed (0.04 * 850) speed deviation (real is 5 but also round misalignment is considered) */

#define UART0_DIVISION(X,Y) (((X) / (Y)) + ((((X) % (Y)) > ((Y) / 2)) ? 1 : 0)) /* division with rounding */
#define UART0_DIVIDED_UART0_CLOCK UART0_DIVISION(F_CPU,8) /* UART0_CLOCK / 8 divided and rounded (used in more macros) */

#define UBRR0_VALUE (UART0_DIVISION(UART0_DIVIDED_UART0_CLOCK , UART0_BAUDRATE) - 1) /* register value */
#define UART0_REAL_BAUDRATE_MUL (UART0_DIVISION(UART0_DIVIDED_UART0_CLOCK * UBRR0_MULTIPLIER , UBRR0_VALUE + 1)) /* real baudrate multiplied and rounded */
#define UART0_REAL_BAUDRATE (UART0_DIVISION(UART0_DIVIDED_UART0_CLOCK , UBRR0_VALUE + 1)) /* real baudrate rounded */
#define uart0_absoluteValue(x) (((x) > 0) ? (x) : -(x)) /* absolute value of number */
#if (UART0_DIVISION(uart0_absoluteValue(UART0_REAL_BAUDRATE_MUL - (UART0_BAUDRATE * UBRR0_MULTIPLIER)) , UART0_BAUDRATE)) > UART0_ALLOWED_DEVIATION_MUL
  #error Required UART0 speed cannot be reached under current circumstances. Change UART0_BAUDRATE or UART0_CLOCK definition to overcome this error.
#endif


#if UART0_TX_BUFFER_SIZE > 254
typedef uint16_t UART0_txPointer_t;
#else
typedef uint8_t UART0_txPointer_t;
#endif

#if !UART_RX_CALLBACK_SUPPORT
#if UART0_RX_BUFFER_SIZE > 254
typedef uint16_t UART0_rxPointer_t;
#else
typedef uint8_t UART0_rxPointer_t;
#endif
#endif


#if UART0_RX_CALLBACK_SUPPORT || DOXYGEN
/**
 * \ingroup uart0_read
 * Function type called by callback when \c UART0_RX_CALLBACK_SUPPORT is enabled and 1 byte is received.
 *
 * \note This definition is available only if callback support is defined.
 */
typedef void (* UART0_rxCallback_t)(uint8_t byte);
#endif

#if !UART0_RX_CALLBACK_SUPPORT
extern volatile bool UART0_receivedByteFlag;
#endif

void UART0_init(void);

#if UART0_RX_CALLBACK_SUPPORT
void UART0_setCallback(UART0_rxCallback_t callback);
#endif

uint16_t UART0_readData(uint8_t *buffer, uint16_t countToReceive);
#define UART0_READBYTE_NO_DATA -1

uint16_t UART0_writeData(const uint8_t *buffer, uint16_t countToSend);

int16_t UART0_readByte(void);

bool UART0_writeByte(uint8_t byte);


#endif
